//
//! Raw packet handling, encryption, decryption, parsing

/// Fenrir Connection id
/// 0 is special as it represents the handshake
/// Connection IDs are to be considered u64 little endian
#[derive(Debug, Copy, Clone, PartialEq)]
pub enum ConnectionID {
    /// Connection id 0 represent the handshake
    Handshake,
    /// Non-zero id can represent any connection
    ID(::core::num::NonZeroU64),
}

impl ConnectionID {
    /// New random service ID
    pub fn new_rand(rand: &::ring::rand::SystemRandom) -> Self {
        use ::ring::rand::SecureRandom;
        let mut raw = [0; 8];
        let mut num = 0;
        while num == 0 {
            rand.fill(&mut raw);
            num = u64::from_le_bytes(raw);
        }
        #[allow(unsafe_code)]
        unsafe {
            ConnectionID::ID(::core::num::NonZeroU64::new_unchecked(num))
        }
    }
    /// Quick check to know if this is an handshake
    pub fn is_handshake(&self) -> bool {
        *self == ConnectionID::Handshake
    }
    /// length if the connection ID in bytes
    pub const fn len() -> usize {
        8
    }
    /// write the ID to a buffer
    pub fn serialize(&self, out: &mut [u8]) {
        assert!(out.len() == 8, "out buffer must be 8 bytes");
        match self {
            ConnectionID::Handshake => out[..].copy_from_slice(&[0; 8]),
            ConnectionID::ID(id) => {
                out[..].copy_from_slice(&id.get().to_le_bytes())
            }
        }
    }
}

impl From<u64> for ConnectionID {
    fn from(raw: u64) -> Self {
        if raw == 0 {
            ConnectionID::Handshake
        } else {
            #[allow(unsafe_code)]
            unsafe {
                ConnectionID::ID(::core::num::NonZeroU64::new_unchecked(raw))
            }
        }
    }
}

impl From<[u8; 8]> for ConnectionID {
    fn from(raw: [u8; 8]) -> Self {
        let raw_u64 = u64::from_le_bytes(raw);
        raw_u64.into()
    }
}

/// Fenrir packet structure
#[derive(Debug, Copy, Clone)]
pub struct Packet {
    /// Id of the Fenrir connection.
    id: ConnectionID,
}
